from dataclasses import dataclass
from typing import Dict, Any, Optional


@dataclass
class RuntimeRequirements:
    num_cpus: int
    # This number is in MB
    max_memory_mb: int
    # timeout in seconds
    timeout: int

    @classmethod
    def default(
        cls,
    ):

        return cls(
            num_cpus=1,
            max_memory_mb=1024,
            timeout=600,
        )

    @classmethod
    def from_dict(cls, indict: Dict[str, Any]):
        num_cpus = indict["num_cpus"]
        max_memory_mb = indict["max_memory_mb"]
        timeout = indict["timeout"]

        return cls(num_cpus=num_cpus, max_memory_mb=max_memory_mb, timeout=timeout)



from enum import Enum
from dataclasses import dataclass
from pathlib import Path
import pathlib
from typing import Optional, Callable

from pysmelt import rc


class SmeltPathType(Enum):
    Absolute = 1
    SmeltRootRelative = 2
    SmeltCommandDefRelative = 3


@dataclass(frozen=True)
class SmeltPath:
    path: str

    @property
    def as_str(self):
        return self.path

    @classmethod
    def from_str(cls, path: str):
        return cls(path=path)

    def __str__(self):
        return self.path

    @staticmethod
    def translate(in_path: str) -> str:
        return SmeltPath.from_str(in_path).to_abs_path()

    def to_abs_path(self) -> str:
        smelt_root = rc.SmeltRcHolder.current_smelt_root()
        if Path(self.path).is_absolute():
            return f"{self.path}"
        else:
            return f"{smelt_root}/{self.path}"


@dataclass
class SmeltFilePath:
    """
    A file object in smelt can either be absolute or defined relative to the command_def path

    for instance, if we have an output "foo"

    """

    # TODO Need to clarify what the distinction is between SmeltPath and SmeltFilePath.

    path_type: SmeltPathType
    path: str

    @classmethod
    def abs_path(cls, path: str):
        return cls(path_type=SmeltPathType.Absolute, path=path)

    @classmethod
    def from_str(cls, path: str):
        if Path(path).is_absolute():
            path_type = SmeltPathType.Absolute
        else:
            path_type = SmeltPathType.SmeltCommandDefRelative
        return cls(path_type=path_type, path=path)

    def __str__(self):
        return self.path

    def to_abs_path(self, abs_cmd_path: str):
        if self.path_type == SmeltPathType.SmeltRootRelative:
            return f"{abs_cmd_path}/{self.path}"
        elif self.path_type == SmeltPathType.Absolute:
            return f"{self.path}"
        else:
            raise NotImplementedError(f"Unhandled variant {self.path_type}")


# Type of a callback that fetches the contents of a SmeltPathFetcher. This is
# meant for scenarios where the Smelt file data is located somewhere other than
# the local filesystem, and custom logic is needed to fetch it.
SmeltPathFetcher = Callable[[SmeltPath], str]


def local_smelt_path_fetcher(smelt_path: SmeltPath) -> str:
    """The default SmeltPathFetcher, which fetches the smelt path from the local filesystem."""
    abs_path = smelt_path.to_abs_path()
    with open(abs_path, "r") as f:
        return f.read()


@dataclass(frozen=True)
class TempTarget:
    name: str
    file_path: SmeltPath

    @classmethod
    def parse_string_smelt_target(
        cls, raw_target: str, current_file: Optional[str] = None
    ):
        # Split the string on the colon
        if raw_target.startswith("//"):

            parts = raw_target.split(":")

            # Check if the split resulted in exactly two parts
            if len(parts) != 2:
                raise ValueError("TargetRef was formatted incorrectly")

            # The first part is the path
            path = parts[0]

            # Remove the leading '//' from the path
            if path.startswith("//"):
                path = path[2:]

            if pathlib.Path(path).suffix == ".py":
                pass

            # The second part is the target name
            target_name = parts[1]
            path = SmeltPath.from_str(path)
            if not pathlib.Path(path.to_abs_path()).exists:
                raise RuntimeError(f"path does not exist {path.to_abs_path()}")

            return cls(name=target_name, file_path=path)
        else:
            assert current_file
            target_name = raw_target
            return cls(name=target_name, file_path=SmeltPath.from_str(current_file))



from dataclasses import dataclass
from typing import cast
import betterproto
from pysmelt.proto.smelt_telemetry import Event, SmeltError, SmeltErrorType


class ClientErr(RuntimeError):
    """
    Error that gets thrown when we  receive a client error message
    """

    pass


class SmeltErrEx(RuntimeError):
    """
    Error that gets thrown when we receive an internal SMELT error
    """

    pass


@dataclass
class SmeltErrorHandler:
    """
    Simple subscriber that tells us if we're done executing
    """

    def process_message(self, message: Event):
        (variant, event_payload) = betterproto.which_one_of(message, "et")
        if variant == "error":
            error = cast(SmeltError, event_payload)
            if error.sig == SmeltErrorType.CLIENT_ERROR:
                raise ClientErr(error.error_payload)
            if error.sig == SmeltErrorType.INTERNAL_ERROR:
                raise SmeltErrEx(error.error_payload)
        else:
            pass



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Callable, Dict, Optional, cast
import betterproto
from pysmelt.proto.smelt_telemetry import (
    CommandEvent,
    CommandFinished,
    CommandStdout,
    Event,
)


StdoutSink = Callable[[str], None]


@dataclass
class StdoutPrinter:
    """
    Simple subscriber that prints the stdout for a single command
    """

    command_ref: str
    sink: StdoutSink

    def process_message(self, message: Event):
        (variant, event_payload) = betterproto.which_one_of(message, "et")
        if variant == "command":
            event_payload = cast(CommandEvent, event_payload)
            (command_name, command_payload) = betterproto.which_one_of(
                event_payload, "CommandVariant"
            )
            if command_name == "stdout":
                command_payload = cast(CommandStdout, command_payload)
                self.sink(command_payload.output)

            else:
                pass
        else:
            pass



_template_fn = f"""
from dataclasses import dataclass, field
from pysmelt.interfaces import Target, SmeltFilePath, SmeltTargetType, TargetRef
from typing import List, Dict


@dataclass
class your_rule(Target):
    deps: List[TargetRef] = field(default_factory=list)

    def gen_script(self) -> List[str]:
        raise NotImplementedError

    def get_dependencies(self) -> List[TargetRef]:
        raise NotImplementedError
        return self.deps
"""


def create_rule_target_from_template(filename: str):
    with open(filename, "w") as f:
        f.write(_template_fn)



from dataclasses import dataclass, field
from functools import partial
from pysmelt.interfaces import Target, SmeltFilePath, SmeltTargetType, TargetRef
from typing import Any, List, Dict, Optional

from pysmelt.interfaces.runtime import RuntimeRequirements


@dataclass
class raw_bash(Target):
    """
    Simple target for embedding raw bash commands in Smelt

    Environment variables available, to all targets are:
        * ${SMELT_ROOT}: the root of the smelt-workspace -- by default, this will be ${GIT_ROOT}
        * ${TARGET_ROOT}: the working space of the current command -- it will be ${SMELT_ROOT}/smelt-out/${COMMAND_NAME}
    """

    cmds: List[str] = field(default_factory=list)
    deps: List[TargetRef] = field(default_factory=list)
    outputs: Dict[str, str] = field(default_factory=dict)
    debug_cmds: Optional[List[str]] = None
    rebuild_cmds: Optional[List[str]] = None
    num_cpus: Optional[int] = None
    timeout: Optional[int] = None
    mem_usage: Optional[int] = None

    def gen_script(self) -> List[str]:
        return self.cmds

    def gen_rerun_script(self) -> Optional[List[str]]:
        return self.debug_cmds

    def gen_rebuild_script(self) -> Optional[List[str]]:
        return self.rebuild_cmds

    def get_dependencies(self) -> List[TargetRef]:
        return self.deps

    def get_outputs(
        self,
    ) -> Dict[str, str]:
        return self.outputs

    def runtime_requirements(
        self,
    ) -> RuntimeRequirements:
        rr = RuntimeRequirements.default()
        if self.num_cpus:
            rr.num_cpus = self.num_cpus
        if self.timeout:
            rr.timeout = self.timeout
        return rr


@dataclass
class raw_bash_build(raw_bash):
    @staticmethod
    def rule_type() -> SmeltTargetType:
        return SmeltTargetType.Build


@dataclass
class test_group(Target):
    """
    Target for oragnizing tests
    """

    tests: List[TargetRef] = field(default_factory=list)

    def gen_script(self) -> List[str]:
        return [f"echo {test}" for test in self.tests]

    def get_dependencies(self) -> List[TargetRef]:
        return self.tests



from dataclasses import dataclass, field, asdict
from abc import ABC
from enum import Enum
from functools import partial
from typing import Any, List, Dict, Literal, Optional, TypedDict
from pysmelt.interfaces.command import Command
from pysmelt.interfaces.runtime import RuntimeRequirements
from pysmelt.interfaces.paths import SmeltFilePath
from pysmelt.rc import SmeltRcHolder
from pysmelt.tracker import ImportTracker, try_get_target


class SmeltTargetType(Enum):
    Test = "test"
    Stimulus = "stimulus"
    Build = "build"
    ## not to be used by end users
    Rebuild = "rebuild"
    Rerun = "rerun"


TargetRef = str


smelt_target = partial(dataclass, frozen=True)()


NamedFiles = Dict[str, str]
"""
Named files are a collection of names mapping to SmeltPaths

We use str instead of the SmeltPath in the value to keep implimentation simple

"""


@dataclass
class Target(ABC):
    """
    A target is a structure that holds logic to generate a `Command`

    Targets are higher level abstraction to commands -- they allow users to embed "application logic" into targets

    """

    name: str

    @property
    def ws_path(self) -> str:
        return f"$SMELT_ROOT/smelt-out/{self.name}"

    def get_outputs(
        self,
    ) -> NamedFiles:
        return {}

    def gen_script(self) -> List[str]:
        raise NotImplementedError

    def gen_rebuild_script(self) -> Optional[List[str]]:
        return None

    def gen_rerun_script(self) -> Optional[List[str]]:
        return None

    @staticmethod
    def rule_type() -> SmeltTargetType:
        return SmeltTargetType.Test

    def runtime_requirements(
        self,
    ) -> RuntimeRequirements:
        return RuntimeRequirements.default()

    def get_dependencies(
        self,
    ) -> List[TargetRef]:
        """
        Returns the targets that this target depends on
        """
        return []

    def get_dependent_files(
        self,
    ) -> List[str]:
        """
        Returns the files that this target depends on
        """
        return []

    @property
    def as_ref(self) -> TargetRef:
        """
        Currently refs are just the names of each target
        """
        return self.name

    def _default_to_command(self, working_dir: str) -> Command:
        name = self.name
        target_type = self.rule_type().value
        script = self.gen_script()
        runtime = self.runtime_requirements()
        dependencies = self.get_dependencies()
        dependent_files = self.get_dependent_files()
        rerun_command = self.to_rerun_command(working_dir)
        outputs = list(map(lambda path: str(path), self.get_outputs().values()))
        return Command(
            name=name,
            target_type=target_type,
            script=script,
            runtime=runtime,
            dependencies=dependencies,
            dependent_files=dependent_files,
            outputs=outputs,
            working_dir=working_dir,
            on_failure=f"{rerun_command.name}" if rerun_command else None,
        )

    def to_command(self, working_dir: str) -> Command:
        return self._default_to_command(working_dir)

    def to_rerun_command(self, working_dir: str) -> Optional[Command]:
        return self.default_rerun_command(working_dir)

    def to_rebuild_command(self, working_dir: str) -> Optional[Command]:
        return self.default_rebuild_command(working_dir)

    def default_rerun_command(self, working_dir: str) -> Optional[Command]:
        script = self.gen_rerun_script()
        if script:
            name = f"{self.name}--rerun"
            target_type = SmeltTargetType.Rerun.value
            runtime = self.runtime_requirements()
            dependencies = [
                rebuild_command.name
                for dep in (
                    try_get_target(f"{idep}") for idep in self.get_dependencies()
                )
                if dep is not None
                for rebuild_command in [dep.to_rebuild_command(working_dir)]
                if rebuild_command is not None
            ]

            dependent_files = self.get_dependent_files()
            outputs = list(map(lambda path: str(path), self.get_outputs().values()))
            return Command(
                name=name,
                target_type=target_type,
                script=script,
                runtime=runtime,
                dependencies=dependencies,
                dependent_files=dependent_files,
                outputs=outputs,
                working_dir=working_dir,
            )

    def default_rebuild_command(self, working_dir: str) -> Optional[Command]:
        script = self.gen_rebuild_script()
        if script:
            name = f"{self.name}--rebuild"
            target_type = SmeltTargetType.Rebuild.value
            runtime = self.runtime_requirements()
            dependencies = [
                rebuild_command.name
                for dep in (
                    try_get_target(f"{idep}") for idep in self.get_dependencies()
                )
                if dep is not None
                for rebuild_command in [dep.to_rebuild_command(working_dir)]
                if rebuild_command is not None
            ]

            dependent_files = self.get_dependent_files()
            outputs = list(map(lambda path: str(path), self.get_outputs().values()))
            return Command(
                name=name,
                target_type=target_type,
                script=script,
                runtime=runtime,
                dependencies=dependencies,
                dependent_files=dependent_files,
                outputs=outputs,
                working_dir=working_dir,
            )

    def __post_init__(self):
        pass
